// $Id: basefile.cc,v 1.9 2005/11/23 14:11:14 troein Exp $
//
// BioArray Software Environment (BASE) - homepage http://base.thep.lu.se/
// Copyright (C) 2002,2003 Lao Saal, Carl Troein, Johan Vallon-Christersson
//
// This file is part of BASE.
//
// BASE is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// BASE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
#ifdef HAVE_CONFIG_H
	#include <config.h>
#endif

#include <cstring>
#include <iostream>

#include <basefile.h>

namespace BASE
{

// Should be in a different file
void split(const std::string &s, std::vector<std::string> &v,
	size_t maxp, char c)
{
	v.clear();
	size_t cp = 0, oldcp = 0;
	while((!maxp || v.size() < maxp) &&
		(cp = s.find(c, oldcp)) != std::string::npos)
	{
		v.push_back(s.substr(oldcp, cp - oldcp));
		oldcp = cp + 1;
	}
	v.push_back(s.substr(oldcp));
}



// Should be in a different file
std::istream &getline(std::istream &i, std::vector<char> &v, char delim)
{
	if(v.size() < 32)
		v.resize(32);
	v[0] = 0;
	size_t vp = 0;

	if(i.peek() == delim)
	{
		i.get();
		return i;
	}
	while(i.get(&v[vp], v.size() - vp, delim))
	{
		if(i.peek() == delim)
		{
			i.get();
			break;
		}
		else
		{
			vp = v.size() - 1;
			v.resize(v.size() * 2);
		}
	}
	return i;
}


const double BaseFileReader::dnan = sqrt(-1.0);
const float BaseFileReader::fnan = float(sqrt(-1.0));

BaseFileSection::BaseFileSection()
{
	startLine = dataStartLine = -1; 
}

bool BaseFileSection::findIntOpt(const std::string &name, int &ret)
{
	header_t::iterator i = headers.find(name);
	if(i == headers.end())
		return false;
	ret = atoi((*i).second.c_str());
	return true;
}

bool BaseFileSection::findStringOpt(const std::string &name, std::string &ret)
{
	header_t::iterator i = headers.find(name);
	if(i == headers.end())
		return false;
	ret = (*i).second;
	return true;
}

bool BaseFileSection::findFieldList(const std::string &name, FieldList &mapping)
{
	header_t::iterator i = headers.find(name);
	if(i == headers.end())
		return false;

	mapping.clear();
	char buf[(*i).second.size()+1];
	strcpy(buf, (*i).second.c_str());
	char *tok = strtok(buf, "\t");
	while(tok)
	{
		int s = mapping.size();
		mapping[std::string(tok)] = s;
		tok = strtok(NULL, "\t");
	}
	return true;
}

bool BaseFileSection::findFieldPos(const std::string &name, int &ret,
	FieldList &mapping)
{
	FieldList::iterator i = mapping.find(name);
	if(i == mapping.end())
		return false;
	ret = (*i).second;
	return true;
}


#ifdef HAVE_PUBSETBUF
const int BaseFileReader::inbuflen = 0x8000;
#else
const int BaseFileReader::inbuflen = 0;
#endif

const char *BaseFileReader::errtexts[] = 
{
	"No error", "File I/O error", "Bad file format", "Too many fields",
	"Too few fields", "Wrong file type", "Bad section specification",
	"File seek error", "invalid error"
};

BaseFileReader::BaseFileReader(std::istream &s, bool changebuf) :
	str(s)
{
	line = 0;
	_error = EOR;
	dataEnd = true;
	_firstSection = true;
#ifdef HAVE_PUBSETBUF
	if(changebuf)
	{
		inbuf.resize(inbuflen);
		s.rdbuf()->pubsetbuf(&inbuf[0], inbuflen);
	}
#endif
}

BaseFileReader::~BaseFileReader()
{
#ifdef HAVE_PUBSETBUF
	if(inbuf.size() > 0)
		str.rdbuf()->pubsetbuf(NULL, 0);
#endif
}

bool BaseFileReader::readLine()
{
	line++;

	getline(str, buf);
	if(!str.good())
	{
		_error = str.bad() ? FILEERR : EOR;
		dataEnd = true;
		return false;
	}
	if(!buf[0])
		dataEnd = true;
	return true;
}

bool BaseFileReader::readHeadersNew(BaseFileSection::header_t &nameValue)
{
	dataEnd = false;
	nameValue.clear();
	if(_firstSection)
	{
		if(!readLine())
			return false;
		char *bp = &buf[0];
		if(*bp == '#') bp++;
		if(strcmp(bp, "BASEfile"))
		{
			_error = BADFORMAT;
			return false;
		}
		_firstSection = false;
	}
	if(!readLine())
		return false;
	while(!buf[0])
	{
		if(!readLine())
			return false;
	}
	char *bp = &buf[0];
	if(*bp == '#') bp++;
	while(buf[0] && strcmp(bp, "%"))
	{
		char *tab = strpbrk(bp, "\t");
		if(tab)
		{
			*tab = 0;
			nameValue[std::string(bp)] = std::string(&tab[1]);
		}
		else nameValue[std::string(bp)] = "";
		if(!readLine())
			return false;
		bp = &buf[0];
		if(*bp == '#') bp++;
	}
	dataEnd = false;
	return true;
}

bool BaseFileReader::skipData()
{
	if(!dataEnd)
	{
		if(!readLine())
			return false;
		while(!dataEnd)
		{
			if(!readLine())
				return false;
		}
	}
	return true;
}

bool BaseFileReader::readSection(BaseFileSection &sect)
{
	skipData();
	sect.startLine = line + 1;

	if(!readHeadersNew(sect.headers))
		return false;
	sect.dataStartLine = line + 1;
	BaseFileSection::header_t::iterator i = sect.headers.find("section");
	if(i == sect.headers.end() || (*i).second.empty())
	{
		_error = BADSECTION;
		return false;
	}
	sect.type = (*i).second;
	return true;
}

bool BaseFileReader::readLines(std::vector<std::string> &vec)
{
	dataEnd = false;	// Kludge for compatibility.
	if(!readLine()) return false;
	while(!dataEnd)
	{
		vec.push_back(&buf[0]);
		if(!readLine())
			return false;
	}
	return true;
}

bool BaseFileReader::tellPos(SeekPosition &spos)
{
	spos.line = line;
	spos.pos = str.tellg();
	if(spos.pos < 0)
	{
		_error = SEEKERR;
		return false;
	}
	return true;
}
bool BaseFileReader::seekPos(const SeekPosition &spos)
{
	str.clear();
	str.seekg(spos.pos);
	if(str.bad() || str.tellg() != spos.pos)
	{
		_error = SEEKERR;
		return false;
	}
	line = spos.line;
	dataEnd = false;
	return true;
}

};	// end namespace BASE

